var utils = require('sk/node/utils'),
getMime = utils.getMime,
Url   	    = require('url'),
ViewFactory		= require('./view').ViewFactory;  
EventEmitter = require('events').EventEmitter; 
Cache		 = require('../io/cache').Cache;       
querystring = require('querystring'), 
session = require('./session'),
MemSessionManager = session.MemSessionManager;




exports.RouteRequest = RouteRequest;   
exports.Router = Router;

function RouteRequest(uri, path, router, request, data, fileExtension, commands)
{     
	var self = this; 

	//emits changes which get streamed back to the client, or MULTIPLE clients
	//if the current request is shared
	var emitter = new EventEmitter();  

	//the commands being invoked
	commands = commands || [];
	
	//the URI requested by the client
	this.uri = uri;

	//the file path (items that aren't in the router)
	this.path = path;

	this.request = request;

	//the router that passes the client request to the routes
	this.router = router;

	//the method invoked: POST, GET, DELETE, etc.
	this.method = request.method;

	//the raw request from the client. this shouldn't be handled directly.
	this.client = request;

	//the header files to send back to the client
	this.headers = {};

	//the data built from the URI, or the query string
	this.data = data;      

	//the host being invked
	this.host = request.headers.host;   

	//the current route we're invoking
	this.currentRoute = commands[commands.length-1];

	this.lastRoute = commands[0];

	//the file extension associated with the URI. default = html
	this.fileExtension = fileExtension;       

	//TRUE if the headers have been sent. 
	this.headersSent = false; 

	//OK by default
	this.statusCode = 200;
	
	//private data usually set by pass-thru routes
	this.privateData = {};
	
	var streamStarted = false;
	

	//moves onto the next URI
	this.next = function(data)
	{   
		
		if(commands.length == 0)                      
		return null;      

		data = data || {};

		for(var i in data)
		{
			self.data[i] = data[i];
		}

		var nextRoute = self.currentRoute = commands.pop();  

		if(nextRoute.init)
		{
			nextRoute.init(self);
		}
		                                          
		nextRoute.callback(self);  
		
		//there's a race condition where if the current route hasn't yet used
		//the post data, we need to listen for it so it doesn't get dumped out of 
		//memory. This CERTAINLY isn't ideal, but it's only good option I can think of :/
		if(self.method == 'POST' && !streamStarted)
		{
			cacheStreamData();
		}
		
		return true;
	}

	this.on = function(type,callback)
	{
		emitter.addListener(type,callback);
	} 

	var cachedStreamData   = [],
		cachedDataFinished = false,
		caching = false;
	
	function cacheStreamData()
	{
		if(caching) return;
		caching = true;
		
		self.streamData(function(data)
		{
			cachedStreamData.push(data);
		}
		,function(data)
		{
			if(data) cachedStreamData.push(data);
			cachedDataFinished = true;
		});
	}   
	                    

	//parses the query string. this is an OPTIONAL method since the data sent to the server
	//could be a multi-part form data / raw post / etc.
	this.parseQueryString = function(callback)
	{                      
		if(self.method == 'GET') return callback(self.data);
		                      
		var qs = "";

		this.streamData(function(chunk)
		{                                       
			qs += chunk;
		},
		function(chunk)
		{                        
			var data = querystring.parse(qs); 
			
			for(var i in data)
			{
				self.data[i] = unescape(data[i]);
			}  
			
			callback(self.data);
		});   
	}
	
	
	this.streamData = function(onData,onEnd)
	{
		
		cachedStreamData.forEach(function(chunk)
		{
			onData(chunk);
		});
		
		if(cachedDataFinished)
		{
			return onEnd();
		}
		
		request.on('data',onData);
		request.on('end',onEnd);
	}

	this.send = this.write = function(data)
	{                            
		if(!self.headersSent)
		{
			self.writeHead();
		}

		emitter.emit('data',data);
	}

	this.session = function(callback)
	{
		if(this.sess) return callback(this.sess);
		
		router.getNewSession(this,function(session)
		{
			self.sess = session;
			callback(session);
		})
		
	}
	

	this.redirect = function(url)
	{
		this.statusCode = 302;
		this.headers['Location'] = url;
		this.end();
	}

	this.getHostDirs = function()
	{
		return self.router.hostDirs[self.host] || self.router.hostDirs.$all;
	}    

	this.getRouteUri = function(cmdOrUri, params)
	{
		return router.getRouteUri(cmdOrUri,params);
	}


	this.getRouteUrl = function(cmdOrUri,params)
	{
		return router.getRouteUrl(cmdOrUri,params,this.host);
	}

	this.display = function(templateName,data)
	{
		if(typeof templateName == 'object')
		{
			data= templateName;
			templateName = 'json';
		}

		ViewFactory.display(this,templateName,data);
	}

	this.end = function(data)
	{    
		if(!self.headersSent)
		{
			self.writeHead();
		}

		emitter.emit('end',data);
	}     

	this.writeHead = function(statusCode)
	{
		if(statusCode) this.statusCode = statusCode;
		this.headersSent = true;
		if(this.sess && this.sess.send) self.headers['Set-Cookie'] = this.sess.toString();
		emitter.emit('response',self.headers);
	}      
	                   
}



function Router()
{         
	var self = this;                   
	this.hostsDir		 = '';
	this.routes 		 = {};      
	this.hostDirs 		 = {};//directores for the host
	this.routesByUri 	 = {};
	this.routesByCommand = {};          
	this.hosts			 = []; //list of hosts
	this.sessionManager = new MemSessionManager();
	
	//used to increment the command if it doesn't exist
	var numCommands = 0;

	this.addRouteInfo = function(info,hosts)
	{   
		var types = {get:'GET',post:'POST',OPTIONS:'OPTIONS',all:['OPTIONS','POST','GET']}          
		
		hosts = hosts || self.hosts;     
		  
		for(var i = hosts.length; i--;)
		{
			var host = hosts[i];          
			
		   	for(var j in types)
			{                  
				var type = types[j]; 
				type = type instanceof Array ? type : [type];
                       
				for(var k = type.length; k--;)
				{   
					addRoutes(info[j], host, type[k], info);  
				}
			} 
		}         
	}
	
	this.addHost = function(host, dirs)
	{
     	var newHosts = host instanceof Array ? host : [host];

		for(var i = newHosts.length; i--;)
		{
			var newHost = newHosts[i];
			
			if(self.hosts.indexOf(newHost)+1)
			{
				continue;//exists
			}
			
			self.hosts.push(newHost);
			
			
			self.hostDirs[newHosts] = dirs;
		}

	}     

	this.getNewSession = function(request,callback)
	{
		return this.sessionManager.getSession(request,callback);
	}

	function addRoutes(routes,host,method,info)
	{   
		if(!routes)
		return;

		if(!self.routes[host])
		{
			self.routes[host] 	   = {};
			self.routesByUri[host] = {}
		}

		if(!self.routes[host][method])
		{
			self.routes[host][method] 	   = {}; 
			self.routesByUri[host][method] = {};
		}

		if(routes instanceof Array)
		{
			for(var i in routes)
			{  
				//we clone the object since "all" be used              
				addRoute(utils.clone(routes[i]),host,method,info);
			}
		}
		else
		{
			addRoute(utils.clone(routes),host,method,info);
		}
		reOrganizePassThrus(null,null,method,host);	
	}   

	function addRoute(route, host, method, info)
	{                         

		//the prefix for the uri's
		var prefix     = info.prefix;   

		//the view info where the templates live
		route.viewInfo = info.view || {};

		//the host for this URI
		route.host     = host;

		//called every time a command is ececuted
		route.init     = info.init; 
		
		//secure tells the router when fetching getRouteUrl to use https instead
		route.secure = route.secure != undefined ? route.secure : info.secure;

		var ttl = {};

		if(route.ttl instanceof Object)
		{
			ttl.client = route.ttl.client || -1;
			ttl.server = route.ttl.server || -1;
		}
		else
		{
			ttl.client = ttl.server = route.ttl || -1;
		}


		route.ttl = ttl;


		//the time to live on the client            

		//the uris need to be an array, but the developer has the option to set the uri as a string
		var uris  = route.uri = route.uri instanceof Array ? route.uri : [route.uri];

		//the command for the uri. this is used when calling getUri from the route
		var command = route.command = route.command || ('route'+(numCommands++));


		for(var j in uris)
		{
			var uri = uris[j] = (prefix || '')+uris[j];


			var pieces = uri.split(/\/+/);

			var currentRoute = self.routes[host][method];   

			var ri;

			for(var k in pieces)
			{
				var dir = pieces[k];            

				if(dir == '') continue;

				var key = dir;  

				ri = {};

				var firstChar = dir.substr(0,1);


				//:var(\\w+)*
				if(firstChar == ":")
				{
					key = "$param";    
					/*ri.$argName = dir.substr(1);  

					var regindex = dir.indexOf("(");

					if(regindex > -1)
					{                                    
						var regex = dir.substr(regindex);   


						//if there's a trailing ALL *, the user wants to push the groups as parameters instead of grabbing the first
						//result. e.g: :fullName((\\w+)+(\\w+))* to function(first,last){} 

						var endPiece = regex.substr(regex.lastIndexOf(")"));

						if(endPiece[1] == "*")
						{
							ri.$useAll = true;
						}

						ri.$regex = new RegExp("^"+regex.substr(1,regex.length-endPiece.length-1)+"$");
					}*/       
				}                  

				if(!currentRoute[key])
				{
					currentRoute[key] = ri;
				} 

				currentRoute = currentRoute[key];
			} 
			
			var params = uri.match(/\/:[^\/]+/g);
			var p = {}
			if(params)
			{    
				params.forEach(function(name)
				{                         
					p[name.substr(2)] = 1;
				})   
			}
			currentRoute.$argNames = p;
			if(currentRoute.$info)
			{                       
				throw new Error("A conflicting route between "+uri+" exists.");
			}                                                                      

			currentRoute.$info = route;

			var fixedUri = pieces.join('/');

			if(!self.routesByCommand[host]) self.routesByCommand[host] = {};    
			
			self.routesByCommand[host][command]      = route;
			self.routesByUri[host][method][fixedUri] = route;


		}    
	}

	function reOrganizePassThrus(currentRoutes,parentPassThru,method,host)
	{               
		var routes = currentRoutes || self.routes,
		passThru = routes.$all ? routes.$all.$info : parentPassThru;            

		for(var route in routes)
		{                     

			var routeTree = routes[route],
			currentPassThru;

			if(route == '$info' || 
			route == '$argName' ||
			route == '$argNames' || 
			route == "$regex")
			continue;

			if(routeTree.$all)
			{
				currentPassThru = routeTree.$all.$info;
			}                                              
			else
			{
				currentPassThru = passThru;
			}                              

			reOrganizePassThrus(routeTree,passThru,method,host);

			if(routeTree.$info)
			{
				var info = routeTree.$info;

				if(info.passThru != undefined)
				{                               

					var pn = info.passThru; 

					//if the pass thru is false, then it's specified that NO pass thru methods
					//thould be executed before the route. this is useful for public routes
					if(pn == false)
					{
						info.$passThru = null;
					}
					else
					if(self.routesByCommand[host] && self.routesByCommand[host][pn])
					{
						info.$passThru = self.routesByCommand[host][pn]; 
					}
					else
					if(self.routesByUri[host] && self.routesByUri[host][method])
					{
						info.$passThru = self.routesByUri[host][method][pn];
					}                          
				}
				else
				if(route == '$all')
				{
					info.$passThru = parentPassThru; 
				}                                                                 
				else
				{
					info.$passThru = currentPassThru;
				}

			}
		}

	} 

	this.getRouteUri = function(cmdOrUri,params,host)
	{
		if(cmdOrUri.indexOf('/') > -1)
		{
			return null;
		}
		else
		{

			params = params || [];

			var info = this.getRoute(cmdOrUri,host);

			if(!info)
			return null;

			var uri = info.uri[0];

			while(params.length > 0 && uri.match(/:[^\/]+/))
			{
				uri = uri.replace(/:[^\/]+/,params.shift());
			}

			return uri;
		}
	}

	this.getRoute = function(cmdOrUri,host)
	{
		var route = (this.routesByCommand[host] || this.routesByCommand.$all)[cmdOrUri]; 
		
		if(!route)
		{
			throw new Error("The route "+cmdOrUri+" does not exist")
		}
		
		return route;
	}

	this.getRouteUrl = function(cmdOrUri,params,host)
	{                     
	   	if(!host)
		{
			for(var i in this.routesByCommand)
			{
				if(i != '$all' && this.routesByCommand[i][cmdOrUri])
				{
					host = i;           
					break;
				}
			}
		}       
		
		var route = this.getRoute(cmdOrUri,host);
		
			
		
		return (route.secure?'https':'http')+'://'+(host+'/'+this.getRouteUri(cmdOrUri,params,host)).replace(/\/+/g,'/');
	}

	this.getRouteRequest = function(request)
	{                     
		var url = request.url;
		
		var file = getFileInfo(url),
		host = request.headers.host,
		method = request.method;
		
		
		//request.data = internal
		var data = request.data || {},
		commands = [],
		info = setRouteCommands(file.route, data, host, method, commands),
		routeRequest = new RouteRequest(file.route.join('/'),
		info.path,
		self,
		request,
		data,
		file.extension,
		commands);   


		var q = file.info.query;

		for(var i in q)
		{
			routeRequest.data[i] = q[i];
		}
		

		routeRequest.uriInfo = file.info;
		self.request = routeRequest;
		
		
		return routeRequest;

	}   

	this.getUri = function(command,data)
	{
		var route = this.routesByCommand[command];  

		var uri = route.uri.toString().replace("$all",""); 

		for(var i in data) uri = uri.replace('/\w+/',data[i]);

		return uri;
	}

	function getPassThru(route,host,uriPieces,i)
	{                 
		i = i || 0;

		//if there is no pass thru given for the current route, then check if a pass thru
		//exists in the current stack
		if(route && route.$all)
		{
			route = route.$all;                                  
		}    
		//if everything fails, then set the pass thru route to the first one, which should
		//lead to a 404 page                   
		else
		{
			route = (self.routes[host] || self.routes.$all).GET.$all;
		} 
		return route;
	}

	function setRouteCommands(uriPieces, data, host, method, commands)
	{                       
		var route = (self.routes[host] || self.routes.$all)[method],
			params = [];
			
			
		if(route)
		for(var i = 0, n = uriPieces.length; i < n; i++)
		{            
			var piece = uriPieces[i];  
			
			
			if(route[piece])
			{
				route = route[piece];  
			}                        
			else    
			//if the route doesn't exist, then handle a wild card route
			if(route.$param)
			{
				route = route.$param;
				/*var argName = route.$argName; 

				if(route.$regex)
				{            
					var match = piece.match(route.$regex); 	     
					if(!match) continue;      
					data[argName] = route.$useAll ? match : match[0];

				}
				else
				{
					data[argName] = piece;  
				}*/
				params.push(piece);
			}                      
			else
			{
				//if the route has a pass thru, then use it if we cannot fullfil the request
				if(route.$info && route.$info.$passThru)
				{      
					route = {$info:route.$info.$passThru};
				}                                             
				else
				{
					route = getPassThru(route,host,uriPieces,i);
				}

				break;              
			}
		} 
		
		
		if(!route || !route.$info)
		{
			route = getPassThru(route,host,uriPieces);                                    
		} 
		
		//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		//this shouldn't EVER happen, since /** should always be the initial passthru route
		//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		if(!route) return -1;         

		if(route.$argNames)
		{
			var k = 0,arg;
			for(var l in route.$argNames)
			{
				data[l] = params[k];
				// arg = route.$argNames[j];
				k++;
			}
		}

		//NOTE: it's OKAY to call /info because a pass thru must ALWAYS exist
		//the first route is ALWAYS the one we want to hit last
		var cp = route.$info,
		ttl = cp.ttl;
 
		while(cp)
		{     
			commands.push(cp);
			cp = cp.$passThru;  
		}
		var rest = [];

		for(var j = i, n = uriPieces.length; j < n; j++)
		{
			rest.push(uriPieces[j]);
		} 
		

		return {route:cp,path:rest.join('/')};
	}


	function getFileInfo(url)
	{
		var pathInfo = Url.parse(url,true);   

		//uri might not be present. this will happen if we're visiting the root page or /
		var uriPieces = pathInfo.pathname.match(/[^\/]+/g) || []; 
		
		
		//let's FIX the uri so nothing funny happens such as going beyond the root http directory, which 
		//could spell trouble. Hmmmm access to the entire file system? Need to look out for this shit because it only
		//just occurred to me a few months after this was written >.>
		for(var i = 0; i < uriPieces.length; i++)
		{                  
			if(uriPieces[i] == '..')
			{          
				//if the URI piece is at index 0, then we need to start at index 0, and splice just ONCE. We're
				//at the point where we can't go any further. OTHERWISE ste the inc to one so we take two steps back, 
				//and remove TWO items. 
				var inc = i?1:0;             
				
				//remove the previous directory, and .., or just .. if it's at index 0
				uriPieces.splice(i-inc,inc+1); 
				i-=1+inc;
			}
		}                                    
		
		//the uri MUST of something in it. If there is NO uri, then we hit the root of the site
		if(!uriPieces.length) uriPieces.push(''); 
		                             

		//the last index is needed for the file name
		var lastIndex = uriPieces.length-1;     

		//the file name favico.ico
		var fileName = uriPieces[lastIndex];

		//the file pieces [favico,ico]
		var filePieces = fileName.split('.');  

		var fileExt = null;//'html';

		if(uriPieces[0] == '')
		{
			uriPieces.pop();
		}
		else
		{
			if(filePieces.length > 1)
			{
				fileExt = filePieces.pop(); 
			}

			//set the last uri piece to the basename of the file pieces so we have a clean route
			uriPieces[lastIndex] = fileName = filePieces.join('.');
		}
  
		//if the file extension is present, then 
		return {name:fileName, extension:fileExt, route:uriPieces,info:pathInfo};
	}     
	
	
	function handleRequest(request, callbacks)
	{                     
		var r = self.getRouteRequest(request);  
		
		//HMMMMM, maybe this should stay in the View controller vs here eh??
		r.headers = {'Content-Type': getMime(r.fileExtension || 'html')};

		//this chunk is used to allow ajax applications to load data from the server
		//NOTE: that any ajax app makes TWO requests if there are custom headers. The first is to see
		//what headers can be sent, and the second is to actually LOAD the data. the second time around, the origin
		//is omitted, so we provide a wildcard to allow any server to talk to clove
		r.headers['Access-Control-Allow-Origin'] =  request.headers.origin;
		             
		//is ajax
		if(request.headers['access-control-request-headers'])
		{                                      
			//i believe the timestamp is used so the access control is cached client side, omitting the need
			//to hit the server on every request
			r.headers['Date'] = (new Date()).toGMTString();
			
			//allow the headers that were sent
			r.headers['Access-Control-Allow-Headers'] =  request.headers['access-control-request-headers'];
			
			//allow only the most common methods
			r.headers['Access-Control-Allow-Methods'] = 'GET, POST, OPTIONS';
			
			//allow session data to be passed to the server from ajax. this should really be oauth since this is NOT supported
			//in older browsers
			//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			//from W3C: The string "*" cannot be used for a resource that supports credentials.
			//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			// r.headers['Access-Control-Allow-Credentials'] = 'true';
			
			//tell the client to store the access control for 17000 seconds
			r.headers['Access-Control-Max-Age'] = 17000;       
			
		}
		
		r.on('data',function(chunk)
		{                                        
			// cacher.record('data',chunk); 
			callbacks.data(chunk);
		});

		r.on('end',function(chunk)
		{                                      
			// cacher.end(chunk); 
			callbacks.end(chunk);      
		});

		r.on('response',function(headers)
		{	                                       
			callbacks.response({statusCode:r.statusCode || 200, headers: headers});
		});
		
		var route = r.lastRoute;
		                 
		//temporary
		if(!route) return callbacks.end('fail');
		
		//if the URI is meant to be SSL, then don't allow to be public. Re-direct the url to 
		//the secure url
		
		
		
		if(route.secure && !request.socket.encrypted)
		{ 
			return r.redirect('https://'+r.host+request.url);
		}
		
		if(route.ttl.client > -1)
		{
			//cache for the time to live, OR 30 days
			r.headers['Expires'] = new Date(new Date().getTime() + (route.ttl.client || 3600*24*30)*1000).toGMTString();
		} 
		
		if(callbacks.route) callbacks.route(route);
               
		//execute the next route
		r.next();  

	}
	
	function handleDynamic(request, response)  
	{
		handleRequest(request, {
			data:function(chunk)
			{
				response.write(chunk || '','binary');
			}                                         
			,end:function(chunk)
			{
				response.end(chunk);
			}                       
			,response:function(data)
			{               
				response.writeHead(data.statusCode,data.headers);
			}
		})
	}           
	                        
	//bypaasses handleCacheable after one pass...
	// var dynamicGetters = {};       
	
	var cachedUris = {};
	
	
	function handleCacheable(request, response)
	{   
		var url = request.url;
		                       
		var compress = false;//request.headers['accept-encoding'].indexOf('gzip') > -1;       
		
		Cache.open('http',url+request.method+(compress?'compressed':''),-1,
		function(cacher)
		{
			handleRequest(request,{    
				route:function(route)
				{
					//if the route is NOT shared, AND the time to live is -1, then we can dispose the cacher.
					//NOTE that this doesn't dispose the cacher immediately. It waits until we call cacher.end(chunk);
					if(process.env.DEBUG_MODE || (!route.shared && route.ttl.server == -1))
					{         
						// header('Expires: '.gmdate('D, d M Y H:i:s',time()+$ttl).'GMT');                                     
						                      
					}
					else
					{                                               
						//otherwise, tell the cacher to stick around for a little while
						cacher.ttl = route.ttl.server;   
					}                               
					
					cacher.ttl = route.ttl.server; 
					cacher.compress = compress;
					               
					// cacher.delink();   
				}
				,data:function(chunk)
				{                     
					cacher.record('data',chunk);
				}
				,end:function(chunk)
				{
					cacher.end(chunk);
				}                     
				,response:function(chunk)
				{                        
					cacher.record('response',chunk);
				}
			})
		},function(cacher)
		{
			cacher.on('data',function(chunk)
			{
				response.write(chunk || "",'binary');   
				                          
			});

			cacher.on('end',function(chunk)
			{                                    
				response.end(chunk);  
			});

			cacher.on('response',function(data)
			{                                                
				var ttl = 3600*60*24;     
				                                     
				if(compress) 
				{
					data.headers['Content-Encoding'] = 'gzip';
				}
				              
				data.headers['Expires'] = '17-Jan-2038 19:14:07 GMT';
				// data.headers['Expires'] = new Date(new Date().getTime()+ttl*1000).toGMTString();
				data.headers['Cache-Control'] = 'max-age=29030400, public';
				
				response.writeHead(data.statusCode,data.headers);
			});
		})
	}
	
	this.handleInternalRequest = function(uri, data,  callback)
	{
		if(!callback)
		{
			callback = data;
			data = undefined;
		}
		
		var buffer = [];
		
		var request = {
			url: 'http://localhost/' + uri,
			method: 'GET',
			data: data,
			headers: { 
				host: 'localhost'
			}
		}
		
		
		var response = {
			write: function (data)
			{
				buffer.push(data);
			},
			end: function (data)
			{
				if(data) response.write(data);
				
				callback(buffer.join(''));
			},
			writeHead: function(data){ }
		}
		
		self.handleHTTPRequest(request, response);
	}

	this.handleHTTPRequest = function(request, response)
	{                         
		var url = request.url;  
		
		if(!process.env.DEBUG_MODE && request.method == 'GET')
		{
			handleCacheable(request,response);
		}                                     
		else
		{
			handleDynamic(request, response);
		}
	}
}




